#include "stdafx.h"
#include <math.h>
#include "CImaAdpcm.h"

bool CImaAdpcm::m_bInited = false;

const short CImaAdpcm::imaStepSizeTable[] = {
	7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 19, 21, 23, 25, 28, 31, 34,
	37, 41, 45, 50, 55, 60, 66, 73, 80, 88, 97, 107, 118, 130, 143,
	157, 173, 190, 209, 230, 253, 279, 307, 337, 371, 408, 449, 494,
	544, 598, 658, 724, 796, 876, 963, 1060, 1166, 1282, 1411, 1552,
	1707, 1878, 2066, 2272, 2499, 2749, 3024, 3327, 3660, 4026,
	4428, 4871, 5358, 5894, 6484, 7132, 7845, 8630, 9493, 10442,
	11487, 12635, 13899, 15289, 16818, 18500, 20350, 22385, 24623,
	27086, 29794, 32767
};
unsigned char CImaAdpcm::imaStateAdjustTable[ISSTMAX][8];

CImaAdpcm::CImaAdpcm() : m_curpcm(0),m_curstt(0) {}

void CImaAdpcm::Init()
{
	if(m_bInited) return;
	int i,j,k;
	for (i=0; i<=ISSTMAX; i++) {
		for (j=0; j<8; j++) {
			k = i+((j<4)?-1:(2*j-6));
			CImaAdpcm::imaStateAdjustTable[i][j] = max(0, min(k,ISSTMAX));
		}
	}
	m_bInited = true;
}


int CImaAdpcm::Enc(IMAADPCMBLOCK* out, short* in, int smpls, int vol_db)
{
	if(!m_bInited) Init();

	int pcm, state, adpcm, n=0, mpcm = 0;
	out->firstpcm = pcm = m_curpcm;
	out->firststate = state = m_curstt;

	unsigned char *wrptr = out->data;
	double vol = pow(10.0, vol_db/20.0);

	for(int i=0; i<smpls; i++)
	{
		int c, step, delta;

		// Encode
		delta = (int)(in[i]*vol) -pcm;
		step = imaStepSizeTable[state];

		c = (abs(delta)<<2)/step;

		if (c > 7) c = 7;
		if (delta < 0) adpcm = c | 8;
		else adpcm = c;

		// Decode
		delta = ((c+c+1) * imaStepSizeTable[state]) >> 3;

		state = imaStateAdjustTable[state][c];

		if(adpcm != c){
			pcm -= delta;
			if(pcm<-32768) pcm = -32768;
		}
		else {
			pcm += delta;
			if(pcm>32767) pcm = 32767;
		}

		if(i&1)	*wrptr++ |= adpcm << 4;
		else	*wrptr = adpcm;

		if(abs(pcm)>mpcm) mpcm = pcm;
	}

	int db = (int)(log10(mpcm/32768.0)*40)+96;
	if(db<0) db=0;
	out->maxvol = db;

	m_curpcm = pcm;
	m_curstt = state;

	return 0;
}

int CImaAdpcm::Dec(short* out, int smpls, IMAADPCMBLOCK* in)
{
	if(!m_bInited) Init();

	int pcm = in->firstpcm;
	int state = in->firststate;

	unsigned char *rdptr = in->data;

	for(int i=0; i<smpls; i++)
	{
		int adpcm, c, delta;

		// Read
		if(i&1)	adpcm = *rdptr++ >> 4;
		else	adpcm = *rdptr & 0xf;

		// Decode
		c = adpcm & 7;

		delta = ((c+c+1) * imaStepSizeTable[state]) >> 3;
		state = imaStateAdjustTable[state][c];
		
		if(adpcm != c){
			pcm -= delta;
			if(pcm<-32768) pcm = -32768;
		}
		else {
			pcm += delta;
			if(pcm>32767) pcm = 32767;
		}

		// Write
		out[i] = pcm;
	}
	
	return 0;
}
