
#pragma once

template<int nBuffers, int Length> class CWaveIn
{
	BYTE m_buf[nBuffers][Length];
	WAVEHDR m_whdr[nBuffers];

	HWAVEIN m_hwi;
	bool m_bStop;

public:
	CWaveIn() : m_hwi(NULL), m_bStop(true)
	{
		for(int i=0; i<nBuffers; ++i){
			m_whdr[i].dwBufferLength = Length;
			m_whdr[i].dwBytesRecorded = 0;
			m_whdr[i].dwFlags = 0;
			m_whdr[i].dwLoops = 1;
			m_whdr[i].dwUser = 0;
			m_whdr[i].lpData = (LPSTR) m_buf[i];
			m_whdr[i].lpNext = NULL;
			m_whdr[i].reserved = 0;
		}
	}

	MMRESULT Open(HWND hWnd, LPWAVEFORMATEX pwfx)
	{
		Close(); 
		return waveInOpen(&m_hwi, WAVE_MAPPER, pwfx, (DWORD_PTR) hWnd, NULL, CALLBACK_WINDOW);
	}

	MMRESULT Close()
	{
		if(m_hwi == NULL) return 0;
		Stop();
		MMRESULT r = waveInClose(m_hwi);
		return r;
	}

	MMRESULT Start()
	{
		if(m_hwi == NULL) return 0;
		if(m_bStop == false) return -1;

		m_bStop = false;
		for(int i=0; i<nBuffers; ++i){
			waveInPrepareHeader(m_hwi, &m_whdr[i], sizeof(WAVEHDR));
			waveInAddBuffer(m_hwi, &m_whdr[i], sizeof(WAVEHDR));
		}
		return waveInStart(m_hwi);
	}

	MMRESULT Stop()
	{
		if(m_hwi == NULL) return 0;
		m_bStop = true;
		MMRESULT r = waveInReset(m_hwi);
		return r;
	}

	MMRESULT OnData(WPARAM wParam, LPARAM lParam)
	{
		return m_bStop ? waveInUnprepareHeader((HWAVEIN) wParam, (LPWAVEHDR) lParam, sizeof(WAVEHDR))
					   : waveInAddBuffer((HWAVEIN) wParam, (LPWAVEHDR) lParam, sizeof(WAVEHDR));
	}
};

template<int nBuffers, int Length> class CWaveOut
{
	BYTE m_buf[nBuffers][Length];
	WAVEHDR m_whdr[nBuffers];

	HWAVEOUT m_hwo;
	DWORD m_outputtedbufs;
	
public:
	CWaveOut() : m_hwo(NULL), m_outputtedbufs(0)
	{
		for(int i=0; i<nBuffers; ++i){
			m_whdr[i].dwBufferLength = Length;
			m_whdr[i].dwBytesRecorded = 0;
			m_whdr[i].dwFlags = 0;
			m_whdr[i].dwLoops = 1;
			m_whdr[i].dwUser = 0;
			m_whdr[i].lpData = (LPSTR) m_buf[i];
			m_whdr[i].lpNext = NULL;
			m_whdr[i].reserved = 0;
		}
	}
	MMRESULT Open(HWND hWnd, LPWAVEFORMATEX pwfx)
	{
		Close();
		return waveOutOpen(&m_hwo, WAVE_MAPPER, pwfx, (DWORD_PTR) hWnd, NULL, CALLBACK_WINDOW);
	}

	MMRESULT Write(void *data, int datalen)
	{
		if(m_hwo == NULL) return 0;
		int n = m_outputtedbufs++ % nBuffers;
		memcpy(m_whdr[n].lpData, data, min(datalen, Length));
		m_whdr[n].dwBufferLength = min(datalen, Length);
		waveOutPrepareHeader(m_hwo, &m_whdr[n], sizeof(WAVEHDR));
		return waveOutWrite(m_hwo, &m_whdr[n], sizeof(WAVEHDR));
	}

	MMRESULT Close()
	{
		if(m_hwo == NULL) return 0;
		MMRESULT r = waveOutClose(m_hwo);
		m_hwo = NULL;
		return r;
	}

	static MMRESULT OnDone(WPARAM wParam,LPARAM lParam)
	{
		return waveOutUnprepareHeader((HWAVEOUT)wParam, (LPWAVEHDR)lParam, sizeof(WAVEHDR));
	}
};

